--[[
    midi2item by ePi

    使い方
        MIDIアイテムを選択した状態で実行すると,子トラックにノーツの長さ・ピッチに対応した空のアイテムを生成します.
        和音は自動でトラックを分割します.
    
    ライセンス(License)
        The MIT License
        Copyright (C) 2020 ePi

        Permission is hereby granted, free of charge, to any person obtaining a copy ofthis software and associated documentation files (the "Software"),
        to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
        and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,DAMAGES OR OTHER LIABILITY,
        WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    更新履歴
        v1 20/01/07
            公開

        v1.0.1 20/03/17
            エラー処理をxpcallでやるように変更

        v1.1 20/03/24
            同一トラックにあるMIDIアイテムが全く違うトラック群に展開されてしまう問題を修正
            MIDIアイテムのループを行ったとき2回目以降のループがアイテムにならなかった問題を修正
            その他コードの修正

        v1.1.1 20/03/29
            トラックの親子関係が正しく設定できていなかったので修正
            
]]

local R=reaper
local function print(...)
    local t={...}
    for k,v in pairs(t)do t[k]=tostring(v) end
    R.ShowConsoleMsg(table.concat(t,"\t").."\n")
end

local function createNewItem(track,st,ed,pitch)
    local mi=R.AddMediaItemToTrack(track)
    local take=R.AddTakeToMediaItem(mi)
    R.SetMediaItemPosition(mi,st,false)
    R.SetMediaItemLength(mi,ed-st,false)
    R.SetMediaItemTakeInfo_Value(take,"D_PITCH",pitch) 
    return mi
end

local proj=0
R.Undo_BeginBlock2(proj)

xpcall(function()

local tracks={}
for i=0,R.CountSelectedMediaItems(proj)-1 do
    local item=R.GetSelectedMediaItem(proj,i)
    local take=R.GetActiveTake(item)
    if(R.TakeIsMIDI(take))then
        local track=R.GetMediaItemTake_Track(take)
        local idx=R.GetMediaTrackInfo_Value(track,"IP_TRACKNUMBER")
        if(tracks[idx]==nil)then tracks[idx]={} end
        table.insert(
            tracks[idx],
            {
                item=item,
                take=take,
            }
        )
    end
end

if(next(tracks)==nil)then
    error("\nMIDIアイテムが選択されていません.")
end

for track_idx,items in (function(t)
    local kt={}
    for k,v in pairs(t)do table.insert(kt,k) end
    table.sort(kt)
    local i=#kt+1
    local function itr(t,k)
        if(i<1)then return nil end
        i=i-1
        return kt[i],t[kt[i]]
    end
    return itr,t
end)(tracks) do
    local notes={}
    for i,v in ipairs(items) do
        local item_st=R.GetMediaItemInfo_Value(v.item,"D_POSITION")
        local item_ed=item_st+R.GetMediaItemInfo_Value(v.item,"D_LENGTH")
        local take=v.take
        local source_length=R.GetMediaSourceLength(R.GetMediaItemTake_Source(take))
        for nc=0,select(2,R.MIDI_CountEvts(take))-1 do
            local ret,_,_,st,ed,_,pitch,_=R.MIDI_GetNote(take,nc)
            if(ret==false)then goto CONTINUE end
            pitch=pitch-72
            st=R.MIDI_GetProjQNFromPPQPos(take,st)
            ed=R.MIDI_GetProjQNFromPPQPos(take,ed)
            local note={
                R.TimeMap2_QNToTime(proj,st),
                R.TimeMap2_QNToTime(proj,ed),
                pitch
            }

            if(note[2]<=item_st)then
                st=st+source_length
                ed=ed+source_length
                note[1]=st
                note[2]=ed
            elseif(note[1]<item_st)then
                note[1]=item_st
            end
            ::LOOPSOURCE::
            
            if(item_ed<=note[1])then goto CONTINUE end
            if(item_ed<note[2])then note[2]=item_ed table.insert(notes,note) goto CONTINUE end
            table.insert(notes,note)
            st=st+source_length
            ed=ed+source_length
            note={
                R.TimeMap2_QNToTime(proj,st),
                R.TimeMap2_QNToTime(proj,ed),
                pitch
            }
            goto LOOPSOURCE

            ::CONTINUE::
        end
    end
    table.sort(notes,
        function(a,b)
            if(a[1]==b[1])then
                if(a[2]==b[2])then
                    return a[3]>b[3]
                else
                    return a[2]<b[2]
                end
            else
            return a[1]<b[1]
            end
        end
    )

    local tl={{}}
    for j,note in ipairs(notes) do
        local k=1
        repeat
            local tr_ed=tl[k][#tl[k]]
            if(tr_ed==nil)then
                tl[k][1]=note
                break
            elseif(note[1]<tr_ed[2] and tr_ed[1]<note[2])then
                k=k+1
                if(tl[k]==nil)then
                    tl[k]={}
                end
            else
                tl[k][#tl[k]+1]=note
                break
            end
        until #notes<j
    end

    for j,tr in ipairs(tl) do
        local idx=track_idx+j-1
        R.InsertTrackAtIndex(idx,false)
        local track=R.GetTrack(proj,idx)
        for k,note in ipairs(tr) do
            createNewItem(track,table.unpack(note))
        end
    end

    local parent=R.GetTrack(proj,track_idx-1)
    local depth=R.GetMediaTrackInfo_Value(parent,"I_FOLDERDEPTH")
    R.SetMediaTrackInfo_Value(parent,"I_FOLDERDEPTH",1)
    R.SetMediaTrackInfo_Value(R.GetTrack(proj,track_idx+#tl-1),"I_FOLDERDEPTH",depth-1)
end

end,function(err)
    print("エラーが発生しました:\n%s",err)
end)

R.UpdateArrange()

reaper.Undo_EndBlock2(proj,"midi2item work",-1)